"""
Music package

@author: Ronny Andersson (ronny@andersson.tk)
@copyright: (c) 2016 Ronny Andersson
@license: MIT
"""

from . import scales
from . import spn

__all__ = [
           'scales',
           'spn',
           ]
